# -*- coding: utf-8 -*-
import smuthi.spherical_functions as sf
import numpy as np
import sympy.physics.wigner
import sympy
import smuthi.memoizing as memo



def plane_vector_wave_function(x, y, z, kp, alpha, kz, pol):
    r"""Electric field components of plane wave (PVWF).

    .. math::
        \mathbf{\Phi}_j = \exp ( \mathrm{i} \mathbf{k} \cdot \mathbf{r} ) \hat{ \mathbf{e} }_j

    with :math:`\hat{\mathbf{e}}_0` denoting the unit vector in azimuthal direction ('TE' or 's' polarization),
    and :math:`\hat{\mathbf{e}}_1` denoting the unit vector in polar direction ('TM' or 'p' polarization).

    The input arrays should have one of the following dimensions:

        - x,y,z: (N x 1) matrix
        - kp,alpha,kz: (1 x M) matrix
        - Ex, Ey, Ez: (M x N) matrix

    or

        - x,y,z: (M x N) matrix
        - kp,alpha,kz: scalar
        - Ex, Ey, Ez: (M x N) matrix

    Args:
        x (numpy.ndarray): x-coordinate of position where to test the field (length unit)
        y (numpy.ndarray): y-coordinate of position where to test the field
        z (numpy.ndarray): z-coordinate of position where to test the field
        kp (numpy.ndarray): parallel component of k-vector (inverse length unit)
        alpha (numpy.ndarray): azimthal angle of k-vector (rad)
        kz (numpy.ndarray): z-component of k-vector (inverse length unit)
        pol (int): Polarization (0=TE, 1=TM)

    Returns:
        - x-coordinate of PVWF electric field (numpy.ndarray)
        - y-coordinate of PVWF electric field (numpy.ndarray)
        - z-coordinate of PVWF electric field (numpy.ndarray)
    """
    k = np.sqrt(kp**2 + kz**2)
    kx = kp * np.cos(alpha)
    ky = kp * np.sin(alpha)

    scalar_wave = np.exp(1j * (kx * x + ky * y + kz * z))

    if pol == 0:
        Ex = -np.sin(alpha) * scalar_wave
        Ey = np.cos(alpha) * scalar_wave
        Ez = scalar_wave - scalar_wave
    elif pol == 1:
        Ex = np.cos(alpha) * kz / k * scalar_wave
        Ey = np.sin(alpha) * kz / k * scalar_wave
        Ez = -kp / k * scalar_wave
    else:
        raise ValueError('Polarization must be 0 (TE) or 1 (TM)')

    return Ex, Ey, Ez


def spherical_vector_wave_function(x, y, z, k, nu, tau, l, m):
    """Electric field components of spherical vector wave function (SVWF). The conventions are chosen according to
    `A. Doicu, T. Wriedt, and Y. A. Eremin: "Light Scattering by Systems of Particles", Springer-Verlag, 2006
    <https://doi.org/10.1007/978-3-540-33697-6>`_

    Args:
        x (numpy.ndarray):      x-coordinate of position where to test the field (length unit)
        y (numpy.ndarray):      y-coordinate of position where to test the field
        z (numpy.ndarray):      z-coordinate of position where to test the field
        k (float or complex):   wavenumber (inverse length unit)
        nu (int):               1 for regular waves, 3 for outgoing waves
        tau (int):              spherical polarization, 0 for spherical TE and 1 for spherical TM
        l (int):                l=1,... multipole degree (polar quantum number)
        m (int):                m=-l,...,l multipole order (azimuthal quantum number)

    Returns:
        - x-coordinate of SVWF electric field (numpy.ndarray)
        - y-coordinate of SVWF electric field (numpy.ndarray)
        - z-coordinate of SVWF electric field (numpy.ndarray)
    """
    x = np.array(x)
    y = np.array(y)
    z = np.array(z)

    r = np.sqrt(x**2 + y**2 + z**2)
    non0 = np.logical_not(r == 0)
    
    theta = np.zeros(x.shape)
    theta[non0] = np.arccos(z[non0] / r[non0])
    phi = np.arctan2(y, x)

    # unit vector in r-direction

    er_x = np.zeros(x.shape)
    er_y = np.zeros(x.shape)
    er_z = np.ones(x.shape)

    er_x[non0] = x[non0] / r[non0]
    er_y[non0] = y[non0] / r[non0]
    er_z[non0] = z[non0] / r[non0]

    # unit vector in theta-direction
    eth_x = np.cos(theta) * np.cos(phi)
    eth_y = np.cos(theta) * np.sin(phi)
    eth_z = -np.sin(theta)

    # unit vector in phi-direction
    eph_x = -np.sin(phi)
    eph_y = np.cos(phi)
    eph_z = x - x

    cos_thet = np.cos(theta)
    sin_thet = np.sin(theta)
    plm_list, pilm_list, taulm_list = sf.legendre_normalized(cos_thet, sin_thet, l)
    plm = plm_list[l][abs(m)]
    pilm = pilm_list[l][abs(m)]
    taulm = taulm_list[l][abs(m)]

    kr = k * r
    if nu == 1:
        bes = sf.spherical_bessel(l, kr)
        dxxz = sf.dx_xj(l, kr)
        if r == 0:
            bes_kr = (l == 1) / 3
            dxxz_kr =  (l == 1) * 2 / 3
        else:
            bes_kr = bes / kr
            dxxz_kr = dxxz / kr
    elif nu == 3:
        bes = sf.spherical_hankel(l, kr)
        dxxz = sf.dx_xh(l, kr)
        bes_kr = bes / kr
        dxxz_kr = dxxz / kr
    else:
        raise ValueError('nu must be 1 (regular SVWF) or 3 (outgoing SVWF)')

    eimphi = np.exp(1j * m * phi)
    prefac = 1 / np.sqrt(2 * l * (l + 1))
    if tau == 0:
        Ex = prefac * bes * (1j * m * pilm * eth_x - taulm * eph_x) * eimphi
        Ey = prefac * bes * (1j * m * pilm * eth_y - taulm * eph_y) * eimphi
        Ez = prefac * bes * (1j * m * pilm * eth_z - taulm * eph_z) * eimphi
    elif tau == 1:
        Ex = prefac * (l * (l + 1) * bes_kr * plm * er_x + dxxz_kr * (taulm * eth_x + 1j * m * pilm * eph_x)) * eimphi
        Ey = prefac * (l * (l + 1) * bes_kr * plm * er_y + dxxz_kr * (taulm * eth_y + 1j * m * pilm * eph_y)) * eimphi
        Ez = prefac * (l * (l + 1) * bes_kr * plm * er_z + dxxz_kr * (taulm * eth_z + 1j * m * pilm * eph_z)) * eimphi
    else:
        raise ValueError('tau must be 0 (spherical TE) or 1 (spherical TM)')

    return Ex, Ey, Ez


def transformation_coefficients_vwf(tau, l, m, pol, kp=None, kz=None, pilm_list=None, taulm_list=None, dagger=False):
    r"""Transformation coefficients B to expand SVWF in PVWF and vice versa:

    .. math::
        B_{\tau l m,j}(x) = -\frac{1}{\mathrm{i}^{l+1}} \frac{1}{\sqrt{2l(l+1)}} (\mathrm{i} \delta_{j1} + \delta_{j2})
        (\delta_{\tau j} \tau_l^{|m|}(x) + (1-\delta_{\tau j} m \pi_l^{|m|}(x))

    For the definition of the :math:`\tau_l^m` and :math:`\pi_l^m` functions, see
    `A. Doicu, T. Wriedt, and Y. A. Eremin: "Light Scattering by Systems of Particles", Springer-Verlag, 2006
    <https://doi.org/10.1007/978-3-540-33697-6>`_

    Args:
        tau (int):          SVWF polarization, 0 for spherical TE, 1 for spherical TM
        l (int):            l=1,... SVWF multipole degree
        m (int):            m=-l,...,l SVWF multipole order
        pol (int):          PVWF polarization, 0 for TE, 1 for TM
        kp (numpy array):         PVWF in-plane wavenumbers
        kz (numpy array):         complex numpy-array: PVWF out-of-plane wavenumbers
        pilm_list (list):   2D list numpy-arrays: alternatively to kp and kz, pilm and taulm as generated with
                            legendre_normalized can directly be handed
        taulm_list (list):  2D list numpy-arrays: alternatively to kp and kz, pilm and taulm as generated with
                            legendre_normalized can directly be handed
        dagger (bool):      switch on when expanding PVWF in SVWF and off when expanding SVWF in PVWF

    Returns:
        Transformation coefficient as array (size like kp).
    """
    if pilm_list is None:
        k = np.sqrt(kp**2 + kz**2)
        ct = kz / k
        st = kp / k
        plm_list, pilm_list, taulm_list = sf.legendre_normalized(ct, st, l)

    if tau == pol:
        sphfun = taulm_list[l][abs(m)]
    else:
        sphfun = m * pilm_list[l][abs(m)]

    if dagger:
        if pol == 0:
            prefac = -1 / (-1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * (-1j)
        elif pol == 1:
            prefac = -1 / (-1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * 1
        else:
            raise ValueError('pol must be 0 (TE) or 1 (TM)')
    else:
        if pol == 0:
            prefac = -1 / (1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * (1j)
        elif pol ==1:
            prefac = -1 / (1j) ** (l + 1) / np.sqrt(2 * l * (l + 1)) * 1
        else:
            raise ValueError('pol must be 0 (TE) or 1 (TM)')

    B = prefac * sphfun
    return B


def translation_coefficients_svwf(tau1, l1, m1, tau2, l2, m2, k, d, sph_hankel=None, legendre=None, exp_immphi=None):
    r"""Coefficients of the translation operator for the expansion of an outgoing spherical wave in terms of
    regular spherical waves with respect to a different origin:

    .. math::
        \mathbf{\Psi}_{\tau l m}^{(3)}(\mathbf{r} + \mathbf{d} = \sum_{\tau'} \sum_{l'} \sum_{m'}
        A_{\tau l m, \tau' l' m'} (\mathbf{d}) \mathbf{\Psi}_{\tau' l' m'}^{(1)}(\mathbf{r})

    for :math:`|\mathbf{r}|<|\mathbf{d}|`.

    Args:
        tau1 (int):             tau1=0,1: Original wave's spherical polarization
        l1 (int):               l=1,...: Original wave's SVWF multipole degree
        m1 (int):               m=-l,...,l: Original wave's SVWF multipole order
        tau2 (int):             tau2=0,1: Partial wave's spherical polarization
        l2 (int):               l=1,...: Partial wave's SVWF multipole degree
        m2 (int):               m=-l,...,l: Partial wave's SVWF multipole order
        k (float or complex):   wavenumber (inverse length unit)
        d (list):               translation vectors in format [dx, dy, dz] (length unit)
                                dx, dy, dz can be scalars or ndarrays
        sph_hankel (list):      Optional. sph_hankel[i] contains the spherical hankel funciton of degree i, evaluated at
                                k*d where d is the norm of the distance vector(s)
        legendre (list):        Optional. legendre[l][m] contains the legendre function of order l and degree m,
                                evaluated at cos(theta) where theta is the polar angle(s) of the distance vector(s)

    Returns:
        translation coefficient A (complex)

    """
    # spherical coordinates of d:
    dd = np.sqrt(d[0] ** 2 + d[1] ** 2 + d[2] ** 2)

    if exp_immphi is None:
        phid = np.arctan2(d[1], d[0])
        eimph = np.exp(1j * (m1 - m2) * phid)
    else:
        eimph = exp_immphi[m1][m2]

    if sph_hankel is None:
        sph_hankel = [sf.spherical_hankel(n, k * dd) for n in range(l1 + l2 + 1)]

    if legendre is None:
        costthetd = d[2] / dd
        sinthetd = np.sqrt(d[0] ** 2 + d[1] ** 2) / dd
        legendre, _, _ = sf.legendre_normalized(costthetd, sinthetd, l1 + l2)

    A = complex(0)
    for ld in range(abs(l1 - l2), l1 + l2 + 1):
        a5, b5 = ab5_coefficients(l1, m1, l2, m2, ld)
        if tau1 == tau2:
            A += a5 * sph_hankel[ld] * legendre[ld][abs(m1 - m2)]
        else:
            A += b5 * sph_hankel[ld] * legendre[ld][abs(m1 - m2)]
    A = eimph * A
    return A


def translation_coefficients_svwf_out_to_out(tau1, l1, m1, tau2, l2, m2, k, d, sph_bessel=None, legendre=None,
                                             exp_immphi=None):
    r"""Coefficients of the translation operator for the expansion of an outgoing spherical wave in terms of
    outgoing spherical waves with respect to a different origin:

    .. math::
        \mathbf{\Psi}_{\tau l m}^{(3)}(\mathbf{r} + \mathbf{d} = \sum_{\tau'} \sum_{l'} \sum_{m'}
        A_{\tau l m, \tau' l' m'} (\mathbf{d}) \mathbf{\Psi}_{\tau' l' m'}^{(3)}(\mathbf{r})

    for :math:`|\mathbf{r}|>|\mathbf{d}|`.

    Args:
        tau1 (int):             tau1=0,1: Original wave's spherical polarization
        l1 (int):               l=1,...: Original wave's SVWF multipole degree
        m1 (int):               m=-l,...,l: Original wave's SVWF multipole order
        tau2 (int):             tau2=0,1: Partial wave's spherical polarization
        l2 (int):               l=1,...: Partial wave's SVWF multipole degree
        m2 (int):               m=-l,...,l: Partial wave's SVWF multipole order
        k (float or complex):   wavenumber (inverse length unit)
        d (list):               translation vectors in format [dx, dy, dz] (length unit)
                                dx, dy, dz can be scalars or ndarrays
        sph_bessel (list):      Optional. sph_bessel[i] contains the spherical Bessel funciton of degree i, evaluated at
                                k*d where d is the norm of the distance vector(s)
        legendre (list):        Optional. legendre[l][m] contains the legendre function of order l and degree m,
                                evaluated at cos(theta) where theta is the polar angle(s) of the distance vector(s)

    Returns:
        translation coefficient A (complex)
    """
    # spherical coordinates of d:
    dd = np.sqrt(d[0] ** 2 + d[1] ** 2 + d[2] ** 2)

    if exp_immphi is None:
        phid = np.arctan2(d[1], d[0])
        eimph = np.exp(1j * (m1 - m2) * phid)
    else:
        eimph = exp_immphi[m1][m2]

    if sph_bessel is None:
        sph_bessel = [sf.spherical_bessel(n, k * dd) for n in range(l1 + l2 + 1)]

    if legendre is None:
        costthetd = d[2] / dd
        sinthetd = np.sqrt(d[0] ** 2 + d[1] ** 2) / dd
        legendre, _, _ = sf.legendre_normalized(costthetd, sinthetd, l1 + l2)

    A = complex(0), complex(0)
    for ld in range(abs(l1 - l2), l1 + l2 + 1):
        a5, b5 = ab5_coefficients(l1, m1, l2, m2, ld)
        if tau1==tau2:
            A += a5 * sph_bessel[ld] * legendre[ld][abs(m1 - m2)]
        else:
            A += b5 * sph_bessel[ld] * legendre[ld][abs(m1 - m2)]
    A = eimph * A
    return A


@memo.Memoize
def ab5_coefficients(l1, m1, l2, m2, p, symbolic=False):
    """a5 and b5 are the coefficients used in the evaluation of the SVWF translation
    operator. Their computation is based on the sympy.physics.wigner package and is performed with symbolic numbers.

    Args:
        l1 (int):           l=1,...: Original wave's SVWF multipole degree
        m1 (int):           m=-l,...,l: Original wave's SVWF multipole order
        l2 (int):           l=1,...: Partial wave's SVWF multipole degree
        m2 (int):           m=-l,...,l: Partial wave's SVWF multipole order
        p (int):            p parameter
        symbolic (bool):    If True, symbolic numbers are returned. Otherwise, complex.

    Returns:
        A tuple (a5, b5) where a5 and b5 are symbolic or complex.
    """
    jfac = sympy.I ** (abs(m1 - m2) - abs(m1) - abs(m2) + l2 - l1 + p) * (-1) ** (m1 - m2)
    fac1 = sympy.sqrt((2 * l1 + 1) * (2 * l2 + 1) / sympy.S(2 * l1 * (l1 + 1) * l2 * (l2 + 1)))
    fac2a = (l1 * (l1 + 1) + l2 * (l2 + 1) - p * (p + 1)) * sympy.sqrt(2 * p + 1)
    fac2b = sympy.sqrt((l1 + l2 + 1 + p) * (l1 + l2 + 1 - p) * (p + l1 - l2) * (p - l1 + l2) * (2 * p + 1))
    wig1 = sympy.physics.wigner.wigner_3j(l1, l2, p, m1, -m2, -(m1 - m2))
    wig2a = sympy.physics.wigner.wigner_3j(l1, l2, p, 0, 0, 0)
    wig2b = sympy.physics.wigner.wigner_3j(l1, l2, p - 1, 0, 0, 0)

    if symbolic:
        a = jfac * fac1 * fac2a * wig1 * wig2a
        b = jfac * fac1 * fac2b * wig1 * wig2b
    else:
        a = complex(jfac * fac1 * fac2a * wig1 * wig2a)
        b = complex(jfac * fac1 * fac2b * wig1 * wig2b)
    return a, b





# -*- coding: utf-8 -*-
import numpy as np

default_k_parallel = None
default_azimuthal_angles = np.arange(0, 361, 1, dtype=float) * np.pi / 180
default_polar_angles = np.arange(0, 181, 1, dtype=float) * np.pi / 180


def complex_contour(vacuum_wavelength, neff_waypoints, neff_resolution):
    neff_segments = []
    for i in range(len(neff_waypoints)-1):
        neff_segments.append(neff_waypoints[i] + np.arange(0, 1 + neff_resolution / 2, neff_resolution, dtype=complex) * 
                             (neff_waypoints[i + 1] - neff_waypoints[i]))
    return np.concatenate(neff_segments) * angular_frequency(vacuum_wavelength)


def set_default_k_parallel(vacuum_wavelength, neff_waypoints=None, neff_resolution=1e-2, neff_max=None, neff_imag=0.05):
    if neff_waypoints is None:
        neff_waypoints = (0, 0.8, 0.8-1j*neff_imag, neff_max-1j*neff_imag, neff_max)
    global default_k_parallel
    kpar = complex_contour(vacuum_wavelength, neff_waypoints, neff_resolution)
    default_k_parallel = kpar
 
 
def k_z(k_parallel=None, n_effective=None, k=None, omega=None, vacuum_wavelength=None, refractive_index=None):
    """z-component :math:`k_z=\sqrt{k^2-\kappa^2}` of the wavevector. The branch cut is defined such that the imaginary
    part is not negative. Not all of the arguments need to be specified.
 
    Args:
        k_parallel (numpy ndarray):     In-plane wavenumber :math:`\kappa` (inverse length)
        n_effective (numpy ndarray):    Effective refractive index :math:`n_\mathrm{eff}`
        k (float):                      Wavenumber (inverse length)
        omega (float):                  Angular frequency :math:`\omega` or vacuum wavenumber (inverse length, c=1)
        vacuum_wavelength (float):      Vacuum wavelength :math:`\lambda` (length)
        refractive_index (complex):     Refractive index :math:`n_i` of material
 
    Returns:
        z-component :math:`k_z` of wavenumber with non-negative imaginary part (inverse length)
    """
    if k_parallel is None:
        if omega is None:
            omega = angular_frequency(vacuum_wavelength)
        k_parallel = n_effective * omega
 
    if k is None:
        if omega is None:
            omega = angular_frequency(vacuum_wavelength)
        k = refractive_index * omega
 
    kz = np.sqrt(k ** 2 - k_parallel ** 2 + 0j)
    kz = (kz.imag >= 0) * kz + (kz.imag < 0) * (-kz)  # Branch cut such to prohibit negative imaginary
    return kz


def angular_frequency(vacuum_wavelength):
    """Angular frequency :math:`\omega = 2\pi c / \lambda`

    Args:
        vacuum_wavelength (float): Vacuum wavelength in length unit

    Returns:
        Angular frequency in the units of c=1 (time units=length units). This is at the same time the vacuum wavenumber.
    """
    return 2 * np.pi / vacuum_wavelength


def rotation_matrix(alpha=None, beta=None, gamma=None, euler_angles=None):
    if euler_angles is not None:
        alpha = euler_angles[0]
        beta = euler_angles[1]
        gamma = euler_angles[2]
    rotation_matrix_3 = [[np.cos(gamma), np.sin(gamma), 0], [- np.sin(gamma), np.cos(gamma), 0], [0, 0, 1]]
    rotation_matrix_2 = [[np.cos(beta), 0, - np.sin(beta)], [0, 1, 0], [np.sin(beta), 0, np.cos(beta)]]
    rotation_matrix_1 = [[np.cos(alpha), np.sin(alpha), 0], [- np.sin(alpha), np.cos(alpha), 0], [0, 0, 1]] 
    return np.dot(rotation_matrix_3, np.dot(rotation_matrix_2, rotation_matrix_1))


def vector_rotation(r, alpha=None, beta=None, gamma=None, euler_angles=None):
    return np.dot(rotation_matrix(alpha, beta, gamma, euler_angles), r)


def inverse_vector_rotation(r, alpha=None, beta=None, gamma=None, euler_angles=None):
    return np.dot(np.linalg.inv(rotation_matrix(alpha, beta, gamma, euler_angles)), r)




# -*- coding: utf-8 -*-
import os
from setuptools import setup
from setuptools.command.install import install
import pkg_resources
import os
import subprocess
import sys


class CustomInstallCommand(install):
    """Compile nfmds code."""
    def run(self):
        install.run(self)
        # compile nfmds
        if sys.platform.startswith('linux'):
            nfmds_sources_dirname = pkg_resources.resource_filename('smuthi.nfmds', 'NFM-DS')
            os.chdir(nfmds_sources_dirname + '/TMATSOURCES')
            sys.stdout.write('Compiling sources at ' + nfmds_sources_dirname + ' ...')
            sys.stdout.flush()
            subprocess.call(['gfortran', 'TAXSYM_SMUTHI.f90', '-o', 'TAXSYM_SMUTHI.out'])
            sys.stdout.write(' done.\n')
            sys.stdout.flush()
            

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="SMUTHI",
    version="0.8.1",
    author="Amos Egel",
    author_email="amos.egel@kit.edu",
    url='https://gitlab.com/AmosEgel/smuthi',
    description="Light scattering by multiple particles in thin-film systems",
    long_description=read('README.rst'),
    packages=['smuthi', 'smuthi.nfmds'],
    cmdclass={'install': CustomInstallCommand},
    package_data={'smuthi.nfmds': ['NFM-DS/*.txt', 'NFM-DS/TMATSOURCES/*', 'NFM-DS/TMATFILES/*',
                                   'NFM-DS/INPUTFILES/*.dat', 'NFM-DS/OUTPUTFILES/*'],
                  'smuthi': ['data/*']},
    install_requires=['numpy', 'scipy', 'mpmath', 'matplotlib', 'pyyaml', 'argparse', 'imageio', 'sympy', 'tqdm'],
    extras_require={'cuda':  ['PyCuda']},
    entry_points={'console_scripts': ['smuthi = smuthi.__main__:main']},
    license='MIT',
)



"""Check if NFM-DS is installed and otherwise do so."""
import pkg_resources
import os
import subprocess
import sys
from distutils.dir_util import copy_tree
import tempfile

nfmds_sources_dirname = pkg_resources.resource_filename('smuthi.nfmds', 'NFM-DS')
temp_fold = tempfile.TemporaryDirectory(prefix='smuthi_nfmds_')
nfmds_temporary_folder = temp_fold.name
print('Copying NFMDS to temporary directory ', nfmds_temporary_folder)

copy_tree(nfmds_sources_dirname, nfmds_temporary_folder)

if sys.platform.startswith('linux'):
    # does the executable exist?
    if not os.access(nfmds_temporary_folder + '/TMATSOURCES/TAXSYM_SMUTHI.out', os.X_OK):
        cwd = os.getcwd()
        os.chdir(nfmds_temporary_folder + '/TMATSOURCES')
        sys.stdout.write('Compiling sources ...')
        sys.stdout.flush()
        subprocess.call(['gfortran', 'TAXSYM_SMUTHI.f90', '-o', 'TAXSYM_SMUTHI.out'])
        sys.stdout.write(' done.\n')
        sys.stdout.flush()
        os.chdir(cwd)
        


# -*- coding: utf-8 -*-

import numpy as np
import smuthi.spherical_functions
import smuthi.nfmds.t_matrix_axsym as nftaxs
import smuthi.field_expansion as fldex


def mie_coefficient(tau, l, k_medium, k_particle, radius):
    """Return the Mie coefficients of a sphere.

    Args:
        tau         integer: spherical polarization, 0 for spherical TE and 1 for spherical TM
        l           integer: l=1,... multipole degree (polar quantum number)
        k_medium    float or complex: wavenumber in surrounding medium (inverse length unit)
        k_particle  float or complex: wavenumber inside sphere (inverse length unit)
        radius      float: radius of sphere (length unit)

    Returns:
        Mie coefficients as complex
    """
    jlkr_medium = smuthi.spherical_functions.spherical_bessel(l, k_medium * radius)
    jlkr_particle = smuthi.spherical_functions.spherical_bessel(l, k_particle * radius)
    dxxj_medium = smuthi.spherical_functions.dx_xj(l, k_medium * radius)
    dxxj_particle = smuthi.spherical_functions.dx_xj(l, k_particle * radius)

    hlkr_medium = smuthi.spherical_functions.spherical_hankel(l, k_medium * radius)
    dxxh_medium = smuthi.spherical_functions.dx_xh(l, k_medium * radius)

    if tau == 0:
        q = (jlkr_medium * dxxj_particle - jlkr_particle * dxxj_medium) / (jlkr_particle * dxxh_medium - hlkr_medium *
                                                                           dxxj_particle)
    elif tau == 1:
        q = ((k_medium ** 2 * jlkr_medium * dxxj_particle - k_particle ** 2 * jlkr_particle * dxxj_medium) /
             (k_particle ** 2 * jlkr_particle * dxxh_medium - k_medium ** 2 * hlkr_medium * dxxj_particle))
    else:
        raise ValueError('tau must be 0 (spherical TE) or 1 (spherical TM)')

    return q


def t_matrix_sphere(k_medium, k_particle, radius, l_max, m_max):
    """T-matrix of a spherical scattering object.

    Args:
        k_medium (float or complex):            Wavenumber in surrounding medium (inverse length unit)
        k_particle (float or complex):          Wavenumber inside sphere (inverse length unit)
        radius (float):                         Radius of sphere (length unit)
        l_max (int):                            Maximal multipole degree
        m_max (int):                            Maximal multipole order
        blocksize (int):                        Total number of index combinations
        multi_to_single_index_map (function):   A function that maps the SVWF indices (tau, l, m) to a single index

    Returns:
         T-matrix as ndarray
    """
    t = np.zeros((fldex.blocksize(l_max, m_max), fldex.blocksize(l_max, m_max)), dtype=complex)
    for tau in range(2):
        for m in range(-m_max, m_max + 1):
            for l in range(max(1, abs(m)), l_max+1):
                n = fldex.multi_to_single_index(tau, l, m, l_max, m_max)
                t[n, n] = mie_coefficient(tau, l, k_medium, k_particle, radius)
    return t


def t_matrix(vacuum_wavelength, n_medium, particle):
    """Return the T-matrix of a particle.

    Args:
        vacuum_wavelength(float)
        n_medium(float or complex):             Refractive index of surrounding medium
        particle(smuthi.particles.Particle):    Particle object

    Returns:
        T-matrix as ndarray
    """
    if type(particle).__name__ == 'Sphere':
        k_medium = 2 * np.pi / vacuum_wavelength * n_medium
        k_particle = 2 * np.pi / vacuum_wavelength * particle.refractive_index
        radius = particle.radius
        t = t_matrix_sphere(k_medium, k_particle, radius, particle.l_max, particle.m_max)
    elif type(particle).__name__ == 'Spheroid':
        t = nftaxs.tmatrix_spheroid(vacuum_wavelength=vacuum_wavelength, layer_refractive_index=n_medium,
                                    particle_refractive_index=particle.refractive_index,
                                    semi_axis_c=particle.semi_axis_c, semi_axis_a=particle.semi_axis_a,
                                    use_ds=particle.t_matrix_method.get('use discrete sources', True),
                                    nint=particle.t_matrix_method.get('nint', 200),
                                    nrank=particle.t_matrix_method.get('nrank', particle.l_max + 2),
                                    l_max=particle.l_max, m_max=particle.m_max)
        if not particle.euler_angles == [0, 0, 0]:
            t = rotate_t_matrix(t, particle.l_max, particle.m_max, particle.euler_angles, wdsympy=False)
    elif type(particle).__name__ == 'FiniteCylinder':
        t = nftaxs.tmatrix_cylinder(vacuum_wavelength=vacuum_wavelength, layer_refractive_index=n_medium,
                                    particle_refractive_index=particle.refractive_index,
                                    cylinder_height=particle.cylinder_height,
                                    cylinder_radius=particle.cylinder_radius,
                                    use_ds=particle.t_matrix_method.get('use discrete sources', True),
                                    nint=particle.t_matrix_method.get('nint', 200),
                                    nrank=particle.t_matrix_method.get('nrank', particle.l_max + 2),
                                    l_max=particle.l_max, m_max=particle.m_max)
        if not particle.euler_angles == [0, 0, 0]:
            t = rotate_t_matrix(t, particle.l_max, particle.m_max, particle.euler_angles, wdsympy=False)
    else:
        raise ValueError('T-matrix for ' + type(particle).__name__ + ' currently not implemented.')

    return t


def rotate_t_matrix(T, l_max, m_max, euler_angles, wdsympy=False):
    """T-matrix of a rotated particle. 
    
    Args:
        T (numpy.array):        T-matrix 
        l_max (int):            Maximal multipole degree
        m_max (int):            Maximal multipole order
        euler_angles (list):    Euler angles [alpha, beta, gamma] of rotated particle in (zy'z''-convention) in radian
        
    Returns:
        rotated T-matrix (numpy.array)
    """
    
    if euler_angles == [0, 0, 0]:
        return T
    else:
        blocksize = fldex.blocksize(l_max, m_max)
        T_mat_rot = np.zeros([blocksize, blocksize], dtype=complex)
    
        # Doicu, Light Scattering by Systems of Particles, p. 70 (1.115) 
        rot_mat_1 = fldex.block_rotation_matrix_D_svwf(l_max, m_max, -euler_angles[2], -euler_angles[1], 
                                                       -euler_angles[0], wdsympy)
        rot_mat_2 = fldex.block_rotation_matrix_D_svwf(l_max, m_max, euler_angles[0], euler_angles[1], euler_angles[2], 
                                                       wdsympy)     
        T_mat_rot = (np.dot(np.dot(np.transpose(rot_mat_1),T), np.transpose(rot_mat_2)))

        # Mishchenko, Scattering, Absorption and Emission of Light by small Particles, p.120 (5.29)
#       T_rot_matrix = np.dot(np.dot(fldex.rotation_matrix_D(l_max, alpha, beta, gamma), T),
#                             fldex.rotation_matrix_D(l_max, -gamma, -beta, -alpha))

        return T_mat_rot


"""Functions to call NFM-DS (null field method with discrete sources) Fortran code by Doicu et al. for the
generation of T-matrices for non-spherical particles.
The Fortran code comes with the book
A. Doicu, T. Wriedt, and Y. A. Eremin: Light Scattering by Systems of Particles,
1st ed. Berlin, Heidelberg: Springer-Verlag, 2006.
and can also be downloaded from
https://scattport.org/index.php/programs-menu/t-matrix-codes-menu/239-nfm-ds
"""

import smuthi.field_expansion as fldex
import smuthi.nfmds
import os
import subprocess
import numpy as np
import imp
import sys


def tmatrix_spheroid(vacuum_wavelength=None, layer_refractive_index=None, particle_refractive_index=None,
                     semi_axis_c=None, semi_axis_a=None, l_max=None, m_max=None, use_ds=True, nint=None, nrank=None):
    """T-matrix for spheroid, using the TAXSYM.f90 routine from the NFM-DS.

    Args:
        vacuum_wavelength(float)
        layer_refractive_index(float):                  Real refractive index of layer (complex values are not allowed).
        particle_refractive_index(float or complex):    Complex refractive index of spheroid
        semi_axis_c (float):                            Semi axis of spheroid along rotation axis
        semi_axis_a (float):                            Semi axis of spheroid along lateral direction
        l_max (int):                                    Maximal multipole degree
        m_max (int):                                    Maximal multipole order
        use_ds (bool):                                  Flag to switch the use of discrete sources on (True) and
                                                        off (False)
        nint (int):                                     Nint parameter for internal use of NFM-DS (number of points
                                                        along integral). Higher value is more accurate and takes longer
        nrank (int):                                    l_max used internally in NFM-DS

    Returns:
        T-matrix as numpy.ndarray
    """
    filename = 'T_matrix_spheroid.dat'
    taxsym_write_input_spheroid(vacuum_wavelength=vacuum_wavelength, layer_refractive_index=layer_refractive_index,
                                particle_refractive_index=particle_refractive_index, semi_axis_c=semi_axis_c,
                                semi_axis_a=semi_axis_a, use_ds=use_ds, nint=nint, nrank=nrank, filename=filename)
    taxsym_run()
    return taxsym_read_tmatrix(filename=filename, l_max=l_max, m_max=m_max)


def tmatrix_cylinder(vacuum_wavelength=None, layer_refractive_index=None, particle_refractive_index=None,
                     cylinder_height=None, cylinder_radius=None, l_max=None, m_max=None, use_ds=True, nint=None, nrank=None):
    """Return T-matrix for finite cylinder, using the TAXSYM.f90 routine from the NFM-DS.

    Args:
        vacuum_wavelength (float)
        layer_refractive_index (float):                 Real refractive index of layer (complex values are not allowed).
        particle_refractive_index (float or complex):   Complex refractive index of spheroid
        cylinder_height (float):                        Semi axis of spheroid along rotation axis
        cylinder_radius (float):                        Semi axis of spheroid along lateral direction
        l_max (int):                                    Maximal multipole degree
        m_max (int):                                    Maximal multipole order
        use_ds (bool):                                  Flag to switch the use of discrete sources on (True) and
                                                        off (False)
        nint (int):                                     Nint parameter for internal use of NFM-DS (number of points
                                                        along integral). Higher value is more accurate and takes longer
        nrank (int):                                    l_max used internally in NFM-DS

    Returns:
        T-matrix as numpy.ndarray
    """
    filename = 'T_matrix_cylinder.dat'
    taxsym_write_input_cylinder(vacuum_wavelength=vacuum_wavelength, layer_refractive_index=layer_refractive_index,
                                particle_refractive_index=particle_refractive_index, cylinder_height=cylinder_height,
                                cylinder_radius=cylinder_radius, use_ds=use_ds, nint=nint, nrank=nrank,
                                filename=filename)
    taxsym_run()
    return taxsym_read_tmatrix(filename=filename, l_max=l_max, m_max=m_max)


def taxsym_run():
    """Call TAXSYM.f90 routine."""
    cwd = os.getcwd()
    os.chdir(smuthi.nfmds.nfmds_temporary_folder + '/TMATSOURCES')
    with open('../nfmds.log', 'w') as nfmds_log:
        if sys.platform.startswith('win'):
            subprocess.call(['TAXSYM_SMUTHI.exe'], stdout=nfmds_log)
        elif sys.platform.startswith('linux'):
            subprocess.call(['./TAXSYM_SMUTHI.out'], stdout=nfmds_log)
        else:
            raise AssertionError('Platform neither windows nor linux.')
    os.chdir(cwd)

    
def taxsym_write_input_spheroid(vacuum_wavelength=None, layer_refractive_index=None, particle_refractive_index=None,
                                semi_axis_c=None, semi_axis_a=None, use_ds=True, nint=None, nrank=None,
                                filename='T_matrix_spheroid.dat'):
    """Generate input file for the TAXSYM.f90 routine for the simulation of a spheroid.

    Args:
        vacuum_wavelength (float)
        layer_refractive_index (float):                 Real refractive index of layer (complex values are not allowed)
        particle_refractive_index (float or complex):   Complex refractive index of spheroid
        semi_axis_c (float):                            Semi axis of spheroid along rotation axis
        semi_axis_a (float):                            Semi axis of spheroid along lateral direction
        use_ds (bool):                                  Flag to switch the use of discrete sources on (True) and
                                                        off (False)
        nint (int):                                     Nint parameter for internal use of NFM-DS (number of points
                                                        along integral). Higher value is more accurate and takes longer
        nrank (int):                                    l_max used internally in NFM-DS
        filename (str):                                 Name of the file in which the T-matrix is stored
    """
    if layer_refractive_index.imag:
        raise ValueError('Refractive index of surrounding medium  must be real(?)')

    f = open(smuthi.nfmds.nfmds_temporary_folder + '/INPUTFILES/InputAXSYM.dat', 'w')

    f.write('OptProp\n')
    f.write(str(float(vacuum_wavelength)) + '\n')
    f.write(str(float(layer_refractive_index.real)) + '\n')
    f.write('(' + str((particle_refractive_index / layer_refractive_index).real) + ','
            + str((particle_refractive_index / layer_refractive_index).imag) + ')\n')

    f.write('Variables:\n')
    f.write(' - wavelength - wavelength of the incident light in vacuo.\n')
    f.write(' - ind_refMed - refractive index of the ambient medium.\n')
    f.write(' - ind_refRel - relative refractive index of the particle.  \n')

    f.write('\n')
    f.write('MatProp\n')
    f.write('.false.\n')
    f.write('.false.\n')
    f.write('0.1\n')

    f.write(' Variables:\n')
    f.write(' - perfectcond - if perfectcond = t, the particle is perfectly conducting.\n')
    f.write(' - chiral      - if chiral = t, the particle is optical active (chiral).\n')
    f.write(' - kb          - parameter of chirality.\n')

    f.write('\n')
    f.write('GeomProp\n')
    f.write('.false.\n')
    f.write("'../GEOMFILES/prolate.fem'\n")
    f.write('1\n')                          # TypeGeom = 1 for spheroid
    f.write('2\n')                          # Nsurf = 2 for spheroid
    f.write(str(float(semi_axis_c)) + '\n')        # half-height of spheroid
    f.write(str(float(semi_axis_a)) + '\n')       # half-width of spheroid
    f.write('1\n')                          # Nparam=1 for spheroid
    f.write('1.0\n')
    f.write('1.0\n')
    f.write('.false.\n')

    f.write(' Variables:\n')
    f.write(' - FileGeom - if FileGeom = t, the particle geometry is supplied by the \n')
    f.write('              input file FileFEM. \n')
    f.write(' - FileFEM  - name of the file containing the particle geometry. \n')
    f.write(' - TypeGeom - parameter specifying the type of the particle geometry.\n')
    f.write(' - Nsurf	   - number of surface parameters. \n')
    f.write(' - surf(1)  - surface parameter.\n')
    f.write(' - ...  \n')
    f.write(' - surf(Nsurf)  \n')
    f.write(' - Nparam   - number of smooth curves forming the generatrix curve.    \n')
    f.write(' - anorm    - characteristic length of the particle which is used to \n')
    f.write('              normalize the differential scattering cross sections.	 \n')
    f.write(' - Rcirc    - characteristic length of the particle for computing Nrank. \n')
    f.write(' - miror    - if miror = t, the particle is mirror symmetric.	            \n')
    f.write(' NOTE: FOR CHIRAL PARTICLES AND DISTRIBUTED SOURCES SET miror = f.\n')

    f.write('\n')
    f.write('ConvTest\n')
    f.write('.false.\n')
    f.write('.false.\n')

    f.write(' Variables:\n')
    f.write(' - DoConvTest   - if DoConvTest = t, the interactive convergence tests \n')
    f.write('                  over Nint and Nrank are performed.   \n')
    f.write(' - MishConvTest - if MishConvTest = t, estimates of Nint and Nrank are  \n')
    f.write('                  computed with the convergence criterion proposed by \n')
    f.write('                  Mishchenko.        \n')
    f.write(' NOTE: IF THE PARTICLE IS OPTICAL ACTIVE (chiral = t) OR THE PARTICLE\n')
    f.write(' GEOMETRY IS SUPPLIED BY THE FILE FileFEM (FileGeom = t), THE CODE SETS\n')
    f.write(' MishConvTest = f. IN FACT, MISHCHENKO''S CONVERGENCE TEST WILL BE \n')
    f.write(' PERFORMED IF (DS = f AND DoConvTest = t AND chiral = f AND FileGeom = f), \n')
    f.write(' OR (DS = t AND autGenDS = t AND DoConvTest = t AND chiral = f AND \n')
    f.write(' FileGeom = f).   \n')

    f.write('\n')
    f.write('Sources\n')
    if use_ds:
        f.write('.true.\n')
    else:
        f.write('.false.\n')
    f.write('.true.\n')

    f.write(' Variables:\n')
    f.write(' - DS       - if DS = t, distributed sources are used for T-matrix \n')
    f.write('              calculation. 	\n')
    f.write(' - autGenDS - if autGenDS = t, the coordinates of the distributed sources\n')
    f.write('              are generated by the code.\n')
    f.write(' NOTE: IF THE PARTICLE GEOMETRY IS READ FROM FILE (FileGeom = t),\n')
    f.write(' THE CODE SETS autgenDS = f.                                 \n')

    f.write('\n')
    f.write('SourcePosAut\n')
    f.write('.true.\n')
    f.write('0.95\n')

    f.write(' Variables: \n')
    f.write(' - ComplexPlane - if ComplexPlane = t, the distributed sources are placed\n')
    f.write('                  in the complex plane.\n')
    f.write(' - EpsZReIm     - parameter controlling the distribution of the discrete \n')
    f.write('                  sources.\n')
    f.write(' NOTE: THESE VARIABLES MUST BE PROVIDED IF (DS = t AND autgenDS = t).\n')

    f.write('\n')
    f.write('NintNrank\n')
    f.write(str(nint) + '\n')
    f.write(str(nrank) + '\n')

    f.write(' Variables: \n')
    f.write(' - Nint  - number of integration points in computing integrals over the \n')
    f.write('           generatrix curve.\n')
    f.write(' - Nrank - maximum expansion order.  \n')
    f.write(' NOTE: THESE VARIABLES MUST BE PROVIDED IF ((DoConvTest = f) OR \n')
    f.write(' (DS = t AND autgenDS = f)).                  \n')

    f.write('\n')
    f.write('Errors\n')
    f.write('5.e-2\n')
    f.write('5.e-2\n')
    f.write('1.e-2\n')
    f.write('4\n')
    f.write('50\n')
    f.write(' Variables:\n')
    f.write(' - epsNint    - error tolerance for the integration test.    \n')
    f.write(' - epsNrank   - error tolerance for the expansion order test.  \n')
    f.write(' - epsMrank   - error tolerance for the azimuthal order test.  \n')
    f.write(' - dNint	     - number of division points for the integration test \n')
    f.write('                and Mishchenko''s convergence test.   \n')
    f.write(' - dNintMrank - number of division points for azimuthal mode \n')
    f.write('                calculation.\n')

    f.write('\n')
    f.write('Tmat\n')
    f.write("'../TMATFILES/" + filename + "'\n")
    f.write(' Variable:\n')
    f.write(' - FileTmat - name of the file to which the T matrix is written.  \n')

    f.write('\n')
    f.write('PrintProgress\n')
    f.write('.false.\n')
    f.write(' Variable:\n')
    f.write(' - PrnProgress - if PrnProgress = t, the progress of calculation \n')
    f.write('                 is printed. \n')

    f.write(' Comment\n')
    f.write(' This file was generated by the routine smuthi.nfmds_wrappers.taxsym_write_input_spheroid \n')

    f.close()


def taxsym_write_input_cylinder(vacuum_wavelength=None, layer_refractive_index=None, particle_refractive_index=None,
                                cylinder_height=None, cylinder_radius=None, use_ds=True, nint=None, nrank=None,
                                filename='T_matrix_cylinder.dat'):
    """Generate input file for the TAXSYM.f90 routine for the simulation of a finite cylinder.

    Args:
        vacuum_wavelength (float)
        layer_refractive_index (float):                 Real refractive index of layer (complex values are not allowed)
        particle_refractive_index (float or complex):   Complex refractive index of cylinder
        cylinder_height (float):                        Height of cylinder (length unit)
        cylinder_radius (float):                        Radius of cylinder (length unit)
        use_ds (bool):                                  Flag to switch the use of discrete sources on (True) and
                                                        off (False)
        nint (int):                                     Nint parameter for internal use of NFM-DS (number of points
                                                        along integral). Higher value is more accurate and takes longer
        nrank (int):                                    l_max used internally in NFM-DS
        filename (str):                                 Name of the file in which the T-matrix is stored
    """
    if layer_refractive_index.imag:
        raise ValueError('Refractive index of surrounding medium  must be real(?)')

    f = open(smuthi.nfmds.nfmds_temporary_folder + '/INPUTFILES/InputAXSYM.dat', 'w')

    f.write('OptProp\n')
    f.write(str(float(vacuum_wavelength)) + '\n')
    f.write(str(float(layer_refractive_index.real)) + '\n')
    f.write('(' + str((particle_refractive_index / layer_refractive_index).real) + ','
            + str((particle_refractive_index / layer_refractive_index).imag) + ')\n')

    f.write('Variables:\n')
    f.write(' - wavelength - wavelength of the incident light in vacuo.\n')
    f.write(' - ind_refMed - refractive index of the ambient medium.\n')
    f.write(' - ind_refRel - relative refractive index of the particle.  \n')

    f.write('\n')
    f.write('MatProp\n')
    f.write('.false.\n')
    f.write('.false.\n')
    f.write('0.1\n')

    f.write(' Variables:\n')
    f.write(' - perfectcond - if perfectcond = t, the particle is perfectly conducting.\n')
    f.write(' - chiral      - if chiral = t, the particle is optical active (chiral).\n')
    f.write(' - kb          - parameter of chirality.\n')

    f.write('\n')
    f.write('GeomProp\n')
    f.write('.false.\n')
    f.write("'../GEOMFILES/prolate.fem'\n")
    f.write('2\n')                          # TypeGeom=2 for cylinder
    f.write('2\n')                          # Nsurf=2 for cylinder
    f.write(str(float(cylinder_height / 2)) + '\n')        # half-height of cylinder
    f.write(str(float(cylinder_radius)) + '\n')       # radius of cylinder
    f.write('3\n')                          # Nparam=3 for cylinder
    f.write('1.0\n')
    f.write('1.0\n')
    f.write('.false.\n')

    f.write(' Variables:\n')
    f.write(' - FileGeom - if FileGeom = t, the particle geometry is supplied by the \n')
    f.write('              input file FileFEM. \n')
    f.write(' - FileFEM  - name of the file containing the particle geometry. \n')
    f.write(' - TypeGeom - parameter specifying the type of the particle geometry.\n')
    f.write(' - Nsurf	   - number of surface parameters. \n')
    f.write(' - surf(1)  - surface parameter.\n')
    f.write(' - ...  \n')
    f.write(' - surf(Nsurf)  \n')
    f.write(' - Nparam   - number of smooth curves forming the generatrix curve.    \n')
    f.write(' - anorm    - characteristic length of the particle which is used to \n')
    f.write('              normalize the differential scattering cross sections.	 \n')
    f.write(' - Rcirc    - characteristic length of the particle for computing Nrank. \n')
    f.write(' - miror    - if miror = t, the particle is mirror symmetric.	            \n')
    f.write(' NOTE: FOR CHIRAL PARTICLES AND DISTRIBUTED SOURCES SET miror = f.\n')

    f.write('\n')
    f.write('ConvTest\n')
    f.write('.false.\n')
    f.write('.false.\n')

    f.write(' Variables:\n')
    f.write(' - DoConvTest   - if DoConvTest = t, the interactive convergence tests \n')
    f.write('                  over Nint and Nrank are performed.   \n')
    f.write(' - MishConvTest - if MishConvTest = t, estimates of Nint and Nrank are  \n')
    f.write('                  computed with the convergence criterion proposed by \n')
    f.write('                  Mishchenko.        \n')
    f.write(' NOTE: IF THE PARTICLE IS OPTICAL ACTIVE (chiral = t) OR THE PARTICLE\n')
    f.write(' GEOMETRY IS SUPPLIED BY THE FILE FileFEM (FileGeom = t), THE CODE SETS\n')
    f.write(' MishConvTest = f. IN FACT, MISHCHENKO''S CONVERGENCE TEST WILL BE \n')
    f.write(' PERFORMED IF (DS = f AND DoConvTest = t AND chiral = f AND FileGeom = f), \n')
    f.write(' OR (DS = t AND autGenDS = t AND DoConvTest = t AND chiral = f AND \n')
    f.write(' FileGeom = f).   \n')

    f.write('\n')
    f.write('Sources\n')
    if use_ds:
        f.write('.true.\n')
    else:
        f.write('.false.\n')
    f.write('.true.\n')

    f.write(' Variables:\n')
    f.write(' - DS       - if DS = t, distributed sources are used for T-matrix \n')
    f.write('              calculation. 	\n')
    f.write(' - autGenDS - if autGenDS = t, the coordinates of the distributed sources\n')
    f.write('              are generated by the code.\n')
    f.write(' NOTE: IF THE PARTICLE GEOMETRY IS READ FROM FILE (FileGeom = t),\n')
    f.write(' THE CODE SETS autgenDS = f.                                 \n')

    f.write('\n')
    f.write('SourcePosAut\n')
    f.write('.true.\n')
    f.write('0.95\n')

    f.write(' Variables: \n')
    f.write(' - ComplexPlane - if ComplexPlane = t, the distributed sources are placed\n')
    f.write('                  in the complex plane.\n')
    f.write(' - EpsZReIm     - parameter controlling the distribution of the discrete \n')
    f.write('                  sources.\n')
    f.write(' NOTE: THESE VARIABLES MUST BE PROVIDED IF (DS = t AND autgenDS = t).\n')

    f.write('\n')
    f.write('NintNrank\n')
    f.write(str(nint) + '\n')
    f.write(str(nrank) + '\n')

    f.write(' Variables: \n')
    f.write(' - Nint  - number of integration points in computing integrals over the \n')
    f.write('           generatrix curve.\n')
    f.write(' - Nrank - maximum expansion order.  \n')
    f.write(' NOTE: THESE VARIABLES MUST BE PROVIDED IF ((DoConvTest = f) OR \n')
    f.write(' (DS = t AND autgenDS = f)).                  \n')

    f.write('\n')
    f.write('Errors\n')
    f.write('5.e-2\n')
    f.write('5.e-2\n')
    f.write('1.e-2\n')
    f.write('4\n')
    f.write('50\n')
    f.write(' Variables:\n')
    f.write(' - epsNint    - error tolerance for the integration test.    \n')
    f.write(' - epsNrank   - error tolerance for the expansion order test.  \n')
    f.write(' - epsMrank   - error tolerance for the azimuthal order test.  \n')
    f.write(' - dNint	     - number of division points for the integration test \n')
    f.write('                and Mishchenko''s convergence test.   \n')
    f.write(' - dNintMrank - number of division points for azimuthal mode \n')
    f.write('                calculation.\n')

    f.write('\n')
    f.write('Tmat\n')
    f.write("'../TMATFILES/" + filename + "'\n")
    f.write(' Variable:\n')
    f.write(' - FileTmat - name of the file to which the T matrix is written.  \n')

    f.write('\n')
    f.write('PrintProgress\n')
    f.write('.false.\n')
    f.write(' Variable:\n')
    f.write(' - PrnProgress - if PrnProgress = t, the progress of calculation \n')
    f.write('                 is printed. \n')

    f.write(' Comment\n')
    f.write(' This file was generated by the routine smuthi.nfmds_wrappers.taxsym_write_input_spheroid \n')

    f.close()


def taxsym_read_tmatrix(filename, l_max, m_max):
    """Export TAXSYM.f90 output to SMUTHI T-matrix.

    .. todo:: feedback to adapt particle m_max to nfmds m_max

    Args:
        filename (str): Name of the file containing the T-matrix output of TAXSYM.f90
        l_max (int):    Maximal multipole degree
        m_max (int):    Maximal multipole order

    Returns:
        T-matrix as numpy.ndarray
    """

    with open(smuthi.nfmds.nfmds_temporary_folder + '/TMATFILES/Info' + filename, 'r') as info_file:
        info_file_lines = info_file.readlines()

    assert 'The scatterer is an axisymmetric particle' in ' '.join(info_file_lines)

    for line in info_file_lines:
        if line.split()[0:4] == ['-', 'maximum', 'expansion', 'order,']:
            n_rank = int(line.split()[-1][0:-1])

        if line.split()[0:5] == ['-', 'number', 'of', 'azimuthal', 'modes,']:
            m_rank = int(line.split()[-1][0:-1])

    with open(smuthi.nfmds.nfmds_temporary_folder + '/TMATFILES/' + filename, 'r') as tmat_file:
        tmat_lines = tmat_file.readlines()

    t_nfmds = [[]]
    column_index = 0
    for line in tmat_lines[3:]:
        split_line = line.split()
        for i_entry in range(int(len(split_line) / 2)):
            if column_index == 2 * n_rank:
                t_nfmds.append([])
                column_index = 0
            t_nfmds[-1].append(complex(split_line[2 * i_entry]) + 1j * complex(split_line[2 * i_entry + 1]))
            column_index += 1

    t_matrix = np.zeros((fldex.blocksize(l_max, m_max), fldex.blocksize(l_max, m_max)), dtype=complex)

    for m in range(-l_max, l_max + 1):
        n_max_nfmds = n_rank - max(1, abs(m)) + 1
        for tau1 in range(2):
            for l1 in range(max(1, abs(m)), l_max + 1):
                n1 = fldex.multi_to_single_index(tau=tau1, l=l1, m=m, l_max=l_max, m_max=m_max)
                l1_nfmds = l1 - max(1, abs(m))
                n1_nfmds = 2 * n_rank * abs(m) + tau1 * n_max_nfmds + l1_nfmds
                for tau2 in range(2):
                    for l2 in range(max(1, abs(m)), l_max + 1):
                        n2 = fldex.multi_to_single_index(tau=tau2, l=l2, m=m, l_max=l_max, m_max=m_max)
                        l2_nfmds = l2 - max(1, abs(m))
                        n2_nfmds = tau2 * n_max_nfmds + l2_nfmds
                        if abs(m) <= m_rank:
                            if m >= 0:
                                t_matrix[n1, n2] = t_nfmds[n1_nfmds][n2_nfmds]
                            else:
                                t_matrix[n1, n2] = t_nfmds[n1_nfmds][n2_nfmds] * (-1) ** (tau1 + tau2)

    return t_matrix
